/*
 *File name ubitx.h 
 *  The second set of 16 pins on the Raduino's bottom connector are have the three clock outputs and the digital lines to control the rig.
 */ 

 /*  
 *  This assignment is as follows :
 *    Pin   1   2    3    4    5    6    7    8    9    10   11   12   13   14   15   16
 *        +12V+12V CLK0  GND  GND  CLK1 GND  GND  CLK2  GND  D2   D3   D4   D5   D6   D7  
 *  These too are flexible with what you may do with them, for the Raduino, we use them to :
 *  - TX_RX line : Switches between Transmit and Receive after sensing the PTT or the Morse keyer
 *  - CW_KEY line : turns on the carrier for CW
 */

#define TX_RX (7)
#define CW_TONE (6)
#define TX_LPF_A (5)
#define TX_LPF_B (4)
#define TX_LPF_C (3)
#define CW_KEY (2)

/**
 * These are the indices where these user changeable settings are stored  in the EEPROM
 */
#define MASTER_CAL   0
#define LSB_CAL      4
#define USB_CAL      8
#define SIDE_TONE   12
//these are ID's of the VFO's as well as their offset into the EEPROM storage, don't change these 'magic' values
#define VFO_A       16
#define VFO_B       20
#define CW_SIDETONE 24
#define CW_SPEED    28
#define FREQ_INCR   32
#define KEY_TYPE    36
#define MODE        40

/**
 * The uBITX is an upconversion transceiver. The first IF is at 45 MHz.
 * The first IF frequency is not exactly at 45 MHz but about 5 khz lower,
 * this shift is due to the loading on the 45 MHz crystal filter by the matching
 * L-network used on it's either sides.
 * The first oscillator works between 48 MHz and 75 MHz. The signal is subtracted
 * from the first oscillator to arrive at 45 MHz IF. Thus, it is inverted : LSB becomes USB
 * and USB becomes LSB.
 * The second IF of 12 MHz has a ladder crystal filter. If a second oscillator is used at 
 * 57 MHz, the signal is subtracted FROM the oscillator, inverting a second time, and arrives 
 * at the 12 MHz ladder filter thus double inversion, keeps the sidebands as they originally were.
 * If the second oscillator is at 33 MHz, the oscillator is subtracted from the signal, 
 * thus keeping the signal's sidebands inverted. The USB will become LSB.
 * We use this technique to switch sidebands. This is to avoid placing the lsbCarrier close to
 * 12 MHz where its fifth harmonic beats with the Arduino's 16 MHz oscillator's fourth harmonic
 */

// the second oscillator should ideally be at 57 MHz, however, the crystal filter's center frequency 
// is shifted down a little due to the loading from the impedance matching L-networks on either sides
#define SECOND_OSC_USB (56995000l)
#define SECOND_OSC_LSB (32995000l) 
//these are the two default USB and LSB frequencies. The best frequencies depend upon your individual taste and filter shape
#define INIT_USB_FREQ   (11996500l)
// limits the tuning and working range of the uBITX between 3 MHz and 30 MHz
#define LOWEST_FREQ  (3000000l)
#define HIGHEST_FREQ (30000000l)

//we directly generate the CW by programming the Si5351 to the cw tx frequency, hence, both are different modes
//these are the parameters passed to startTx
#define TX_SSB 0
#define TX_CW 1

#define CW_TIMEOUT (350l) 


/**
 * We need to carefully pick assignment of pin for various purposes.
 * There are two sets of completely programmable pins on the Raduino.
 * First, on the top of the board, in line with the LCD connector is an 8-pin connector
 * that is largely meant for analog inputs and front-panel control. It has a regulated 5v output,
 * ground and six pins. Each of these six pins can be individually programmed 
 * either as an analog input, a digital input or a digital output. 
 * The pins are assigned as follows (left to right, display facing you): 
 *      Pin 1 (Violet), A7, KEYER (DASH)
 *      Pin 2 (Blue),   A6, KEYER (DOT)
 *      Pin 3 (Green), +5v 
 *      Pin 4 (Yellow), GND
 *      Pin 5 (Orange), A3, PTT
 *      Pin 6 (Red),    A2, F BUTTON
 *      Pin 7 (Brown),  A1, ENC B
 *      Pin 8 (Black),  A0, ENC A
 *Note: A5, A4 are wired to the Si5351 as I2C interface 
 *       *     
 * Though, this can be assigned anyway, for this application of the Arduino, we will make the following
 * assignment
 * A2 will connect to the PTT line, which is the usually a part of the MIC connector
 * A3 is connected to a push button that can momentarily ground this line. This will be used for RIT/Bandswitching, etc.
 * A6 is connected to the DOT lead from the paddle
 * A7 is connected to the DASH lead from the paddle
  */

#define ENC_A (A1)
#define ENC_B (A0)
#define FBUTTON (A2)


#define PTT   (A3)
#define ANALOG_DOT  (A6)
#define DIGITAL_DASH (A3)


//////////////////////////////////////////////

#define MODE_SSB 0
#define MODE_CW  1

#define BtnPressTime 30
#define SplPressTime 120
#define SavPressTime 360
/////////////////////////  keyer

#define DIT_L 0x01 // DIT latch
#define DAH_L 0x02 // DAH latch
#define DIT_PROC 0x04 // DIT is being processed
#define PDLSWAP 0x08 // 0 for normal, 1 for swap
#define IAMBICB 0x10 // 0 for Iambic A, 1 for Iambic B
enum KSTYPE {IDLE, CHK_DIT, CHK_DAH, KEYED_PREP, KEYED, INTER_ELEMENT };


//////////////////////////////////////////////


