/**
 * File name ubitx_keyer.cpp
 * CW Keyer
 * 
 * The CW keyer handles either a straight key or an iambic / paddle key.  These are now "Menu" selected
 * They use 2 analog lines - A6 for DOT Paddle  plus A3 for DASH Paddle/PTT/Handkey
 * Both A6 requires an external 4.7K pullup resistor to +5V 
 * 
 * The key types now must be selected in a "Menu" which will be defined in another module.
 * 
 * Generating CW
 * The CW is cleanly generated by unbalancing the front-end mixer
 * and putting the local oscillator directly at the CW transmit frequency.
 * The sidetone, generated by the Arduino is injected into the volume control
 */
#include <Arduino.h>
#include "ubitx.h"

extern void stopTx(void);
extern void startTx(byte txMode);
extern void updateDisplay(void);
extern void printLine3(char *c);
extern void printLine4(char *c);
extern void printLine2(char *c);

extern unsigned long sideTone;
extern int cwSpeed;
extern volatile bool inTx;
extern bool Iambic_Key;
extern volatile int ubitx_mode;

extern volatile unsigned char keyerControl;
extern volatile unsigned char keyerState;


volatile bool BitTimer = false;            //it is set to 1 if timing paddle bit times
volatile bool keyDown = false;             //in cw mode, denotes the carrier is being transmitted
volatile uint8_t Last_Bits = 0xFF;;

// in milliseconds, this is the parameter that determines how long the TX will hold between CW key downs

//we store the last paddle's character 
//to alternatively send dots and dashes 
//when both are simultaneously pressed

volatile bool Dot_in_Progress = false;
volatile unsigned long  Dot_Timer_Count = 0;
volatile bool Dash_in_Progress = false;
volatile unsigned long  Dash_Timer_Count = 0;
volatile bool Inter_Bit_in_Progress = false;
volatile unsigned long  Inter_Bit_Timer_Count = 0;
volatile bool Turn_Off_Carrier_in_Progress = false;
volatile unsigned long  Turn_Off_Carrier_Timer_Count = 0;



/**
 * Starts transmitting the carrier with the sidetone
 * It assumes that we have called cwTxStart and not called cwTxStop
 * each time it is called, the cwTimeOut is pushed further into the future
 */
void cwKeydown(void){
  keyDown = 1;                  //tracks the CW_KEY
  tone(CW_TONE, (int)sideTone); 
  digitalWrite(CW_KEY, 1);     
}

/**
 * Stops the CW carrier transmission along with the sidetone
 * Pushes the cwTimeout further into the future
 */
void cwKeyUp(void){
  keyDown = 0;    //tracks the CW_KEY
  noTone(CW_TONE);
  digitalWrite(CW_KEY, 0);    
}



void update_PaddleLatch() {
  if (analogRead(ANALOG_DOT) < 600 ) keyerControl |= DIT_L;
  if (!digitalRead(DIGITAL_DASH)  ) keyerControl |= DAH_L;
}

//////////////////////////////////////////////////////////////////////////////////////////
//   interupt  handlers

////   timers
ISR(TIMER1_OVF_vect) {
 bool continue_loop = true;

  
  if( (Turn_Off_Carrier_in_Progress)&&(Turn_Off_Carrier_Timer_Count > 0) ){
    if( digitalRead( DIGITAL_DASH )){
      Turn_Off_Carrier_Timer_Count = Turn_Off_Carrier_Timer_Count - 1;
      if( Turn_Off_Carrier_Timer_Count <= 0 ){
        Turn_Off_Carrier_in_Progress = false;
        Turn_Off_Carrier_Timer_Count = 0;
        stopTx();
      }
    }
  }

if( (Iambic_Key) && (ubitx_mode == MODE_CW) ){

  if( (Dot_in_Progress)&&(Dot_Timer_Count > 0) ){
    if( !inTx ){
      keyDown = 0;
      startTx(TX_CW);
    }
    if( keyDown == 0 ) cwKeydown();
    Dot_Timer_Count = Dot_Timer_Count - 1;
    if( Dot_Timer_Count <= 0 ){
      Dot_Timer_Count = 0;
      Dot_in_Progress = false;
      cwKeyUp();
    }
  }
  
  
  if( (Inter_Bit_in_Progress)&&(Inter_Bit_Timer_Count > 0) ){
    Inter_Bit_Timer_Count = Inter_Bit_Timer_Count - 1;
    if( Inter_Bit_Timer_Count <= 0 ){
      Inter_Bit_Timer_Count = 0;
      Inter_Bit_in_Progress = false;
    }
  }
  
 while(continue_loop){
  
  switch (keyerState) {
    
    case IDLE:
      if ( (analogRead(ANALOG_DOT)< 600 ) ||
           (!digitalRead(DIGITAL_DASH)) ||
           (keyerControl & 0x03)) {
         update_PaddleLatch();
         keyerState = CHK_DIT;
         Dot_in_Progress = false;
         Dot_Timer_Count = 0;
        Turn_Off_Carrier_Timer_Count = 0;
        Turn_Off_Carrier_in_Progress = false;
      }else{
        continue_loop = false; 
      }
      break;

    case CHK_DIT:
      if (keyerControl & DIT_L) {
        keyerControl |= DIT_PROC;
        keyerState = KEYED_PREP;
        Dot_Timer_Count = cwSpeed;
      }else{
        keyerState = CHK_DAH;
      }
      break;

    case CHK_DAH:
      if (keyerControl & DAH_L) {
        keyerState = KEYED_PREP;
        Dot_Timer_Count = cwSpeed*3;
      }else{
        continue_loop = false; 
        keyerState = IDLE;
      }
      break;

    case KEYED_PREP:
      keyerControl &= ~(DIT_L + DAH_L); // clear both paddle latch bits
      keyerState = KEYED; // next state
      Turn_Off_Carrier_Timer_Count = 0;
      Turn_Off_Carrier_in_Progress = false;
      Dot_in_Progress = true; 
      break;

    case KEYED:
      if (Dot_in_Progress == false) { // are we at end of key down ?
        Inter_Bit_in_Progress = true; 
        Inter_Bit_Timer_Count = cwSpeed;
        keyerState = INTER_ELEMENT; // next state
      }else if (keyerControl & IAMBICB) {
        update_PaddleLatch(); // early paddle latch in Iambic B mode
        continue_loop = false; 
      }else continue_loop = false; 
      break;

    case INTER_ELEMENT:
      // Insert time between dits/dahs
      update_PaddleLatch(); // latch paddle state
      if (Inter_Bit_in_Progress == false) { // are we at end of inter-space ?
        Turn_Off_Carrier_Timer_Count = CW_TIMEOUT;
        Turn_Off_Carrier_in_Progress = true;
        if (keyerControl & DIT_PROC) { // was it a dit or dah ?
          keyerControl &= ~(DIT_L + DIT_PROC); // clear two bits
          keyerState = CHK_DAH; // dit done, check for dah
        }else{
          keyerControl &= ~(DAH_L); // clear dah latch
          keyerState = IDLE; // go idle
        }
      }else continue_loop = false; 
      break;
  }
 }
}
}

void Connect_Interrupts(void){
  cli();
  PCMSK1 |= 0b00001000;    // turn on PTT and Handkey pin PC3, physical A3
  PCICR  |= 0b00000010;    // turn on port c

  TIMSK1 |= (1<<TOIE1);
  sei();
}




ISR(PCINT1_vect){
uint8_t Int_Bits;
static unsigned long last_interrupt_time = 0;
static bool IAmRunning = false;

if( (!Iambic_Key) || (ubitx_mode == MODE_SSB) ){

if(IAmRunning) return; //bail out if we interrupted ourself

  unsigned long interrupt_time = micros();
  // If interrupts come faster than 200ms, assume it's a bounce and ignore
  if( interrupt_time - last_interrupt_time > 10) {
    IAmRunning = true;

    Int_Bits = PINC ^ Last_Bits;
    Last_Bits = PINC;
  
    if( ubitx_mode == MODE_CW ){  
     if( (Int_Bits & 0b00001000)){
        if ( digitalRead(DIGITAL_DASH) == 0 ){
          if (!inTx){
            keyDown = false;
            startTx(TX_CW);
          }
          keyDown = true;                  //tracks the CW_KEY
          tone(CW_TONE, (int)sideTone); 
          digitalWrite(CW_KEY, 1);     
        }else{     
          keyDown = false;    //tracks the CW_KEY
          noTone(CW_TONE);
          digitalWrite(CW_KEY, 0); 
          Turn_Off_Carrier_Timer_Count =  CW_TIMEOUT;  
          Turn_Off_Carrier_in_Progress = true;
        }
      }
     }else if( (Int_Bits & 0b00001000)){
      if (digitalRead(DIGITAL_DASH) == 0 && !inTx){
         startTx(TX_SSB);
       }else{
         stopTx();
       }
     }
    
    last_interrupt_time = interrupt_time;
    IAmRunning = false;
  }

}
}



